
from openai_unified import GPT_calls
import sqlite3
from datetime import datetime
import numpy as np
import asyncio
import base64
import aiosqlite
import os
import shutil

class AsyncImageDescriptionDatabase:
    def __init__(self, db_path='images_async.db'):
        self.db_path = db_path
        self.gpt_client = GPT_calls(use_async=True, json_mode=True)  # Initialize the GPT client for asynchronous operations with JSON mode enabled.
        asyncio.run(self._setup_db())
        # create images folder if it doesn't exist
        if not os.path.exists('images'):
            os.makedirs('images')

    async def _setup_db(self):
        # Connect to the database and create table if it doesn't exist
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()
        cursor.execute("""
                CREATE TABLE IF NOT EXISTS images (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    filename TEXT,
                    description TEXT,
                    embeddings BLOB
                );
            """)
        conn.commit()
        conn.close()

    async def save_image_data(self, filename, description, embeddings):
        # Save image data along with embeddings into the database
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()
        embeddings_blob = sqlite3.Binary(np.array(embeddings).astype(np.float32).tobytes())
        cursor.execute("INSERT INTO images (filename, description, embeddings) VALUES (?, ?, ?)",
                       (filename, description, embeddings_blob))
        conn.commit()
        conn.close()

    async def generate_describe_store_image(self, prompt):
        # Generate, describe and store image asynchronously
        image_response = await self.gpt_client.generate_image_async(prompt, save=False, show=False, quality="hd")
        image = image_response.data[0].b64_json
        image_bytes = base64.b64decode(image)

        image_filename = prompt.split()
        if len(image_filename) > 3:
            # add milliseconds to the filename
            image_filename = '_'.join(image_filename[:3]) + datetime.now().strftime('%Y%m%d%H%M%S%f')
        else:
            image_filename = '_'.join(image_filename) + datetime.now().strftime('%Y%m%d%H%M%S%f')
        filename = f"{image_filename}.png"

        with open(f"images/{filename}", 'wb') as f:
            f.write(image_bytes)

        description = await self.gpt_client.get_gpt_vision_response_async(f"images/{filename}", question="Describe this image in detail as A JSON object. include all relevant details plus colors, themes, setting etc.") # or as a json object
        embeddings = await self.gpt_client.get_embedding_async(description)
        await self.save_image_data(filename, description, embeddings)
        print(f"Image {filename} '.")

    async def run_generation(self, prompt, num_images):
        # Setup the database and run generation tasks
        await self._setup_db()
        tasks = [self.generate_describe_store_image(prompt) for _ in range(num_images)]
        await asyncio.gather(*tasks)

    async def search_images(self, query, top_n=3):
        # Perform a similarity search for the given query
        async with aiosqlite.connect(self.db_path) as conn:
            cursor = await conn.execute("SELECT id, embeddings FROM images")
            images_data = await cursor.fetchall()

        query_embeddings = await self.gpt_client.get_embedding_async(query)
        similarities = []
        for image_id, embeddings_blob in images_data:
            embeddings = np.frombuffer(embeddings_blob, dtype=np.float32)
            similarity = await self.gpt_client.similarity_search_async(query_embeddings, embeddings)
            similarities.append((image_id, similarity))
        
        similarities.sort(key=lambda x: x[1], reverse=True)

        async with aiosqlite.connect(self.db_path) as conn:
            for image_id, similarity in similarities[:top_n]:
                cursor = await conn.execute("SELECT filename, description FROM images WHERE id = ?", (image_id,))
                filename, description = await cursor.fetchone()
                print(f"Filename: {filename}, Description: '{description}', Similarity: {similarity}")
                # create a directory with the first 3 words of the query if the length of the query is greater than 3
                query_words = query.split()
                if len(query_words) > 3:
                    query_dir = '_'.join(query_words[:3])
                else:
                    query_dir = '_'.join(query_words)

                if not os.path.exists(query_dir):
                    os.makedirs(query_dir)

                # copy the images to the directory
                shutil.copy(f'images/{filename}', f'{query_dir}/{filename}')

    async def run_app(self):
        # Main interactive loop for generating and searching images
        await self._setup_db()
        while True:
            choice = input("Choose 'g' to generate images, 's' to find similar images, or 'exit' to quit: ").strip().lower()
            if choice == 'g':
                prompt = input("Enter a prompt for the image generation: ")
                num_images = int(input("Enter the number of images to generate: "))
                await self.run_generation(prompt, num_images)
            elif choice == 's':
                query = input("Enter a search query for similar images: ")
                await self.search_images(query)
            elif choice == 'exit':
                break
            else:
                print("Invalid option. Please choose 'g', 's', or 'exit'.")

def main():
    app = AsyncImageDescriptionDatabase()
    asyncio.run(app.run_app())

if __name__ == "__main__":
    main()
