import os
import json
from openai import OpenAI, AsyncOpenAI
from termcolor import colored
from sklearn.metrics.pairwise import cosine_similarity
import numpy as np
import base64
import re
from exa_py import Exa
from pathlib import Path


class GPT_calls:
    def __init__(self, 
                 name="GPT chat",
                 api_key=None, model="gpt-4-turbo-preview", 
                 max_history_words=10000, 
                 max_words_per_message=None, 
                 json_mode=False, stream=True, 
                 use_async=False,
                 exa_api_key=None,  # Add an argument for the Exa API key
                 perplexity_api_key=None  # Add an argument for the Perplexity API key
                 ):
        
        self.name = name
        self.api_key = api_key or os.getenv("OPENAI_API_KEY") or "YOUR_KEY_HERE"
        self.model = model
        self.history = []
        self.max_history_words = max_history_words
        self.max_words_per_message = max_words_per_message
        self.json_mode = json_mode
        self.stream = stream
        self.use_async = use_async
        self.exa_api_key = exa_api_key or os.getenv("EXA_API_KEY") or "YOUR_KEY_HERE"
        self.exa = Exa(api_key=self.exa_api_key)
        self.perplexity_api_key = perplexity_api_key or os.getenv("PERPLEXITY_API_KEY") or "YOUR_KEY_HERE"

        # print the initialization status
        print(colored(f"{self.name} initialized with json_mode={json_mode}, stream={stream}, use_async={use_async}, max_history_words={max_history_words}, max_words_per_message={max_words_per_message}", "red"))

        if use_async:
            self.client = AsyncOpenAI(api_key=self.api_key)
        if not use_async:
            self.client = OpenAI(api_key=self.api_key)
        if self.perplexity_api_key:
            self.perplexity_client = OpenAI(api_key=self.perplexity_api_key, base_url="https://api.perplexity.ai")

    def add_message(self, role, content):
        if role == "user" and self.max_words_per_message:
            self.history.append({"role": role, "content": str(content) + f" please use {self.max_words_per_message} words or less"})
        elif role == "user" and self.max_words_per_message is None:
            self.history.append({"role": role, "content": str(content)})
        else:
            self.history.append({"role": role, "content": str(content)})

    def print_history_length(self):
        history_length = [len(str(message["content"]).split()) for message in self.history]
        print("\n")
        print(f"current history length is {sum(history_length)} words")

    def clear_history(self):
        self.history.clear()

    def chat(self, question, **kwargs):
        self.add_message("user", question)
        return self.get_response(**kwargs)

    def trim_history(self):
        words_count = sum(len(str(message["content"]).split()) for message in self.history if message["role"] != "system")
        while words_count > self.max_history_words and len(self.history) > 1:
            words_count -= len(self.history[0]["content"].split())
            self.history.pop(1)  # Remove second message because first is always system message

    #############################
    # Async versions of the above
    #############################

    async def add_message_async(self, role, content):
        if role == "user" and self.max_words_per_message:
            self.history.append({"role": role, "content": str(content) + f" please use {self.max_words_per_message} words or less"})
        elif role == "user" and self.max_words_per_message is None:
            self.history.append({"role": role, "content": str(content)})
        else:
            self.history.append({"role": role, "content": str(content)})

    async def clear_history_async(self):
        self.history.clear()
    
    async def chat_async(self, question, **kwargs):
        await self.add_message_async("user", question)
        return await self.get_response_async(**kwargs)
    
    async def trim_history_async(self):
        words_count = sum(len(str(message["content"]).split()) for message in self.history if message["role"] != "system")
        while words_count > self.max_history_words and len(self.history) > 1:
            words_count -= len(self.history[0]["content"].split())
            self.history.pop(1)

    def get_response(self, color="green", should_print=True, **kwargs):
        if self.json_mode:
            response = self.client.chat.completions.create(
                model=self.model,
                messages=self.history,
                stream=True if self.stream else False,
                response_format={"type": "json_object"},
                **kwargs
            )

            if self.stream:
                assistant_response = ""
                for chunk in response:
                    if chunk.choices[0].delta.content: 
                        text_chunk = chunk.choices[0].delta.content 
                        if should_print:
                            print(colored(text_chunk, color), end="", flush=True)
                        assistant_response += str(text_chunk)

                # convert the str to json
                assistant_response = json.loads(assistant_response)

            else:
                assistant_response = json.loads(response.choices[0].message.content)

        else:
            response = self.client.chat.completions.create(
                model=self.model,
                messages=self.history,
                stream=True if self.stream else False,
                **kwargs
            )

            if self.stream:
                assistant_response = ""
                for chunk in response:
                    if chunk.choices[0].delta.content: 
                        text_chunk = chunk.choices[0].delta.content 
                        if should_print:
                            print(colored(text_chunk, "yellow"), end="", flush=True)
                        assistant_response += str(text_chunk)
                print("\n")
            else:
                assistant_response = response.choices[0].message.content       
        
        self.add_message("assistant", str(assistant_response))
        self.trim_history()
        return assistant_response
    
    async def get_response_async(self, color="yellow", should_print=True, **kwargs):
        if self.json_mode:
            response = await self.client.chat.completions.create(
                model=self.model,
                messages=self.history,
                stream=True if self.stream else False,
                response_format={"type": "json_object"},
                **kwargs
            )

            if self.stream:
                assistant_response = ""
                async for chunk in response:
                    if chunk.choices[0].delta.content: 
                        text_chunk = chunk.choices[0].delta.content 
                        if should_print:
                            print(colored(text_chunk, color), end="", flush=True)
                        assistant_response += str(text_chunk)

                # convert the str to json
                assistant_response = json.loads(assistant_response)

            else:
                assistant_response = json.loads(response.choices[0].message.content)

        else:
            response = await self.client.chat.completions.create(
                model=self.model,
                messages=self.history,
                stream=True if self.stream else False,
                **kwargs
            )

            if self.stream:
                assistant_response = ""
                async for chunk in response:
                    if chunk.choices[0].delta.content: 
                        text_chunk = chunk.choices[0].delta.content 
                        print(colored(text_chunk, color), end="", flush=True)
                        assistant_response += str(text_chunk)
                print("\n")
            else:
                assistant_response = response.choices[0].message.content       
        
        await self.add_message_async("assistant", str(assistant_response))
        await self.trim_history_async()
        return assistant_response
    
    #############################
    # GPT Vision methods
    #############################


    def get_gpt_vision_response(self, image_path, question="What’s in this image?",color="yellow", **kwargs):
        # add the question to the history
        self.add_message("user", question)
        with open(image_path, "rb") as image_file:
            base64_image =  base64.b64encode(image_file.read()).decode('utf-8')
       
        response = self.client.chat.completions.create(
            model="gpt-4-vision-preview",
            stream=self.stream,
            messages=[
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": f"{question}. please use {self.max_words_per_message} words or less"} if self.max_words_per_message else {"type": "text", "text": f"{question}"},
                        {
                            "type": "image_url",
                            "image_url": f"data:image/jpeg;base64,{base64_image}",
                        },
                    ],
                }
            ],
            max_tokens=kwargs.pop("max_tokens") if "max_tokens" in kwargs else 500,
            **kwargs
        )
        if self.stream:
            assistant_response = ""
            for chunk in response:
                if chunk.choices[0].delta.content: 
                    text_chunk = chunk.choices[0].delta.content
                    print(colored(text_chunk, color), end="", flush=True)
                    assistant_response += str(text_chunk)

            # convert the str to json
            assistant_response = assistant_response

        else:
            assistant_response = response.choices[0].message.content

        self.add_message("assistant", str(assistant_response))
        self.trim_history()

        return assistant_response
    
    async def get_gpt_vision_response_async(self, image_path, question="What’s in this image?",color="yellow", **kwargs):
        with open(image_path, "rb") as image_file:
            base64_image =  base64.b64encode(image_file.read()).decode('utf-8')
       
        response = await self.client.chat.completions.create(
            model="gpt-4-vision-preview",
            stream=True, # comment this out to see non-streaming
            messages=[
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": f"{question}"},
                        {
                            "type": "image_url",
                            "image_url": f"data:image/jpeg;base64,{base64_image}",
                        },
                    ],
                }
            ],
             max_tokens=kwargs.pop("max_tokens") if "max_tokens" in kwargs else 500,
            **kwargs
        )
        if self.stream:
            assistant_response = ""
            async for chunk in response:
                if chunk.choices[0].delta.content: 
                    text_chunk = chunk.choices[0].delta.content 
                    print(colored(text_chunk, color), end="", flush=True)
                    assistant_response += str(text_chunk)

            # convert the str to json
            assistant_response = assistant_response

        else:
            assistant_response = response.choices[0].message.content

        self.add_message_async("assistant", str(assistant_response))
        self.trim_history_async()

        return assistant_response
    
    #############################
    # Generate image methods
    #############################

    def generate_image(self, prompt, save=True, show=True, **kwargs):
        response = self.client.images.generate(
            model="dall-e-3",
            prompt=prompt,
            response_format="b64_json",
            **kwargs
        )

        # save the image to a file, enumerate if file exists to avoid overwriting
        if save:
            # decode the image
            response = response.data[0].b64_json
            image = base64.b64decode(response)
            file_name = "generated_image.png"
            if os.path.exists(file_name):
                # check how manuy files that start with generated_image
                file_count = len([file for file in os.listdir() if file.startswith("generated_image")])
                file_name = f"generated_image_{file_count}.png"
                with open(file_name, "wb") as f:
                    f.write(image)
                
            with open(file_name, "wb") as f:
                f.write(image)

            print(f"Image saved to {file_name}")
            
        if show:
            # use pillow
            from PIL import Image
            from io import BytesIO
            image = Image.open(BytesIO(image))
            image.show()

        return response
    
    async def generate_image_async(self, prompt, save=True, show=True, **kwargs):
        response = await self.client.images.generate(
            model="dall-e-3",
            prompt=prompt,
            response_format="b64_json",
            **kwargs
        )

        # save the image to a file, enumerate if file exists to avoid overwriting
        if save:
            # decode the image
            response = response.data[0].b64_json
            image = base64.b64decode(response)
            file_name = "generated_image.png"
            if os.path.exists(file_name):
                # check how manuy files that start with generated_image
                file_count = len([file for file in os.listdir() if file.startswith("generated_image")])
                file_name = f"generated_image_{file_count}.png"
                with open(file_name, "wb") as f:
                    f.write(image)

                print(f"Image saved to {file_name}")
                
            with open(file_name, "wb") as f:
                f.write(image)
        if show:
            # use pillow
            from PIL import Image
            from io import BytesIO
            image = Image.open(BytesIO(image))
            image.show()

        return response
    
    #############################
    # Text to speech methods
    #############################

    def text_to_speech(self, text, voice="nova", model="tts-1", save_path="speech.mp3"):
        try:
            response = self.client.audio.speech.create(
                model=model,
                voice=voice,
                input=text
            )
            # Save the audio file
            speech_file_path = Path(save_path)
            response.stream_to_file(speech_file_path)
            print(f"Audio file saved to {speech_file_path}")
        except Exception as e:
            print(f"Error generating speech: {e}")

    async def text_to_speech_async(self, text, voice="nova", model="tts-1", save_path="speech.mp3"):
        try:
            response = await self.client.audio.speech.create(
                model=model,
                voice=voice,
                input=text
            )
            # Save the audio file
            speech_file_path = Path(save_path)
            await response.stream_to_file(speech_file_path)
            print(f"Audio file saved to {speech_file_path}")
        except Exception as e:
            print(f"Error generating speech: {e}")
    

    #############################
    # Embeddings methods
    #############################

    def get_embedding(self, text, model="text-embedding-3-large"):
        response = self.client.embeddings.create(
            input=text,            
            model=model
        )
        return response.data[0].embedding

    async def get_embedding_async(self, text, model="text-embedding-3-large"):
        response = await self.client.embeddings.create(
            input=text,
            model=model
        )
        return response.data[0].embedding
    
    def similarity_search(self, embedding1, embedding2):
        """
        Compute the cosine similarity between two embeddings.
        """
        # Ensure the embeddings are in the correct shape (2D arrays)
        embedding1 = np.array(embedding1).reshape(1, -1)
        embedding2 = np.array(embedding2).reshape(1, -1)

        similarity = cosine_similarity(embedding1, embedding2)

        return similarity[0][0]
    
    #############################
    # Async versions of cosine similarity
    #############################

    async def similarity_search_async(self, embedding1, embedding2):
        """
        Compute the cosine similarity between two embeddings.
        """
        # Ensure the embeddings are in the correct shape (2D arrays)
        embedding1 = np.array(embedding1).reshape(1, -1)
        embedding2 = np.array(embedding2).reshape(1, -1)

        similarity = cosine_similarity(embedding1, embedding2)

        return similarity[0][0]
    

    #############################
    # Exa web search
    #############################

    def exa_search(self, query=None, num_results=3, **kwargs):
        if not self.exa:
            raise ValueError("Exa instance not initialized. Please set the API key first.")
        
        results = self.exa.search_and_contents(query=query, num_results=3, **kwargs)
        return results.results
    
    async def exa_search_async(self, api_key=None, query=None, **kwargs):
        if not self.exa:
            raise ValueError("Exa instance not initialized. Please set the API key first.")
        
        results = await self.exa.search_and_contents(query=query, num_results=3, **kwargs)
        return results.results
    
    #############################
    # Perplexity search
    #############################

    def perplexity_search(self, query=None):
        if not self.perplexity_client:
            raise ValueError("Perplexity instance not initialized. Please set the API key first.")
        
        messages = [
        {
            "role": "system",
            "content": (
                "You are an artificial intelligence assistant and you need to "
                "engage in a helpful, detailed, polite conversation with a user."
            ),
        },
        {
            "role": "user",
            "content": (
                f"{query}"
            ),
        },
    ]

        if not self.stream:
            # chat completion without streaming
            response = self.perplexity_client.chat.completions.create(
                model="pplx-70b-online",
                messages=messages,
            )
            return response.choices[0].message.content

        elif self.stream:
            # chat completion with streaming
            response_stream = self.perplexity_client.chat.completions.create(
                model="pplx-70b-online",
                messages=messages,
                stream=True,
            )
            perplexity_response = ""
            for response in response_stream:
                perplexity_response += response.choices[0].delta.content
                print(colored(response.choices[0].delta.content, "magenta"), end="", flush=True)

            return perplexity_response
        
    #############################
    # moderation methods
    #############################
        
    def moderate_content(self, input_text):
        try:
            response = self.client.moderations.create(input=input_text)
            output = response.results[0]  # Assuming 'results' contains the moderation outcome
            return output.flagged
        except Exception as e:
            print(f"Error during content moderation: {e}")
            return None
        
    async def moderate_content_async(self, input_text):
        try:
            response = await self.client.moderations.create(input=input_text)
            output = response.results[0]  # Assuming 'results' contains the moderation outcome
            return output.flagged
        except Exception as e:
            print(f"Error during content moderation: {e}")
            return None
    
    def help(self):
        # Define color scheme
        title_color = 'green'
        method_name_color = 'blue'
        description_color = 'yellow'
        example_color = 'cyan'

        # Help content with color coding
        help_info_colored = [
            colored("HELP GUIDE: GPT_calls Class\n\n", title_color),
            "This class interacts with the OpenAI API, enabling functionalities like chat, vision tasks, and embeddings.\n\n",
            colored("METHODS:\n", title_color),
            colored("1. __init__(self, api_key=None, model='gpt-4-turbo-preview', max_history_words=5000, max_words_per_message=100, json_mode=False, stream=True, use_async=False)\n", method_name_color),
            colored("- Initializes the GPT_calls object.\n", description_color),
            colored("- Example: gpt_client = GPT_calls(api_key='your_api_key')\n\n", example_color),

            colored("2. add_message(self, role, content)\n", method_name_color),
            colored("- Adds a message to the conversation history.\n", description_color),
            colored("- Example: gpt_client.add_message('user', 'Hello, world!')\n\n", example_color),

            colored("3. print_history_length(self)\n", method_name_color),
            colored("- Prints the current word count of the conversation history.\n", description_color),
            colored("- Example: gpt_client.print_history_length()\n\n", example_color),

            colored("4. clear_history(self)\n", method_name_color),
            colored("- Clears the conversation history.\n", description_color),
            colored("- Example: gpt_client.clear_history()\n\n", example_color),

            colored("5. chat(self, question, **kwargs)\n", method_name_color),
            colored("- Sends a question to the API and retrieves a response.\n", description_color),
            colored("- Example: gpt_client.chat('What is AI?')\n\n", example_color),

            colored("6. get_response(self, color='yellow', should_print=True, **kwargs)\n", method_name_color),
            colored("- Retrieves the response from the API.\n", description_color),
            colored("- Example: gpt_client.get_response()\n\n", example_color),

            colored("7. get_gpt_vision_response(self, image_path, question='What’s in this image?', color='yellow', **kwargs)\n", method_name_color),
            colored("- Analyzes an image and returns a response.\n", description_color),
            colored("- Example: gpt_client.get_gpt_vision_response('path/to/image.jpg')\n\n", example_color),

            colored("8. generate_image(self, prompt, save=True, show=True, **kwargs)\n", method_name_color),
            colored("- Generates an image based on a prompt.\n", description_color),
            colored("- Example: gpt_client.generate_image('A bee')\n\n", example_color),

            colored("9. text_to_speech(self, text, voice='nova', model='tts-1', save_path='speech.mp3')\n", method_name_color),

            colored("10. get_embedding(self, text, model='text-embedding-3-large')\n", method_name_color),
            colored("- Retrieves the embedding for a given text.\n", description_color),
            colored("- Example: embedding = gpt_client.get_embedding('Hello, world!')\n\n", example_color),

            colored("11. similarity_search(self, embedding1, embedding2)\n", method_name_color),
            colored("- Computes cosine similarity between two embeddings.\n", description_color),
            colored("- Example: sim = gpt_client.similarity_search(embedding1, embedding2)\n\n", example_color),

            colored("12. exa_search(self, query=None, num_results=3, **kwargs)\n", method_name_color),
            colored("- Searches the web using the Exa API.\n", description_color),
            colored("- Example: results = gpt_client.exa_search('What is AI?')\n\n", example_color),

            colored("13. perplexity_search(self, query=None)\n", method_name_color),
            colored("- Searches the web using the Perplexity API.\n", description_color),

            colored("14. moderate_content(self, input_text)\n", method_name_color),

            colored("Async versions of these methods are also available.\n", description_color)
        ]

        # Combine the help information into a single string for file writing
        help_info_plain = ''.join([re.sub('\033\\[.*?m', '', line) for line in help_info_colored])


        # Print colored help information
        for line in help_info_colored:
            print(line, end="")

        # Check if help.txt exists, if not, write the plain help info into it
        if not os.path.exists('help.txt'):
            with open('help.txt', 'w') as file:
                file.write(help_info_plain)

        print(colored("\nHelp information is printed. Also saved to 'help.txt' if it didn't exist.", 'magenta'))


if __name__ == "__main__":
    # Example usage
    gpt_client = GPT_calls()

    # Adding system message
    gpt_client.add_message("system", "You are a helpful assistant.")

    # Asking questions
    gpt_client.chat("Who won the world series in 2020?")
    print("\n\n")
    gpt_client.add_message("user", "What is the meaning of life?")

    gpt_client.get_response()

    print(gpt_client.history)
